from __future__ import annotations

from typing import Iterable, NamedTuple

import torch


class _ModuleInfo(NamedTuple):
    attr: str
    path: str
    module: torch.nn.Module
    parent: torch.nn.Module | None


def named_modules_with_parent(model: torch.nn.Module) -> Iterable[_ModuleInfo]:
    for path, module in model.named_modules():
        if path == "":
            assert module is model
            yield _ModuleInfo("", "", model, parent=None)
            continue

        *parent_path, attr = path.split(".")
        parent = model
        for child in parent_path:
            parent = getattr(parent, child)
        assert getattr(parent, attr) is module, (
            "named_modules_with_parents failed to find module in parent by name. "
            "If this occurs, please report a bug."
        )
        yield _ModuleInfo(attr, path, module, parent)



from __future__ import annotations

from collections import Counter
from typing import Any, Iterable, NamedTuple

import torch

import slicetune.nn
import slicetune.utils


def mark_for_training(
    model: torch.nn.Module,
    train_tuners: bool = True,
    train_biases: bool = False,
    train_rest: bool = False,
) -> None:

    for param in model.parameters():
        param.requires_grad_(train_rest)

    for param in get_bias_params(model):
        param.requires_grad_(train_biases)

    for param in get_tuners(model):
        param.requires_grad_(train_tuners)


def fuse(model: torch.nn.Module) -> None:
    """
    Replaces all slicetune layers in the model with equivalent standard layers.
    Enables gradient computation for all parameters.
    """
    # prevent invalidating the iterator by setattr
    module_infos = list(slicetune.utils.named_modules_with_parent(model))
    for module_info in module_infos:
        if isinstance(module_info.module, slicetune.nn.Layer):
            replacement = module_info.module.to_standard()
            setattr(module_info.parent, module_info.attr, replacement)
        module_infos = list(slicetune.utils.named_modules_with_parent(model))
    for param in model.parameters():
        param.requires_grad = True


def get_slicetune_layers(
    model: torch.nn.Module,
) -> Iterable[slicetune.nn.Layer]:
    """
    Returns an iterable of all slicetune layers in the model.
    """
    for module in model.modules():
        if isinstance(module, slicetune.nn.Layer):
            yield module


def get_tuners(
    model: torch.nn.Module,
    only_tuning_params: bool = True,
) -> Iterable[torch.nn.Parameter]:
    """
    Returns an iterable of all slicetuning parameters in the model.
    """
    for module in model.modules():
        if isinstance(module, slicetune.nn.Layer):
            if only_tuning_params:
                params = module.get_tuners()
            else:
                params = module.parameters()

            yield from params


def get_bias_params(model: torch.nn.Module) -> Iterable[torch.nn.Parameter]:
    """
    Returns an iterable of all bias parameters in the model.
    """
    for module in model.modules():
        bias = getattr(module, "bias", None)
        if bias is not None and isinstance(bias, torch.nn.Parameter):
            yield bias


def get_trainable_params(model: torch.nn.Module) -> Iterable[torch.nn.Parameter]:
    """
    Returns an iterable of all trainable parameters in the model.
    """
    for param in model.parameters():
        if param.requires_grad:
            yield param


class SlicetuningStats(NamedTuple):
    params_total: int
    params_total_trainable: int
    params_tuners: int
    params_tuners_trainable: int
    params_biases: int
    params_biases_trainable: int
    layers_count: dict[str, int]

    def _pretty(self, value: Any) -> str:
        if isinstance(value, int):
            return f"{value:,}"
        elif isinstance(value, float):
            return f"{value:.2%}"
        return str(value)

    def _frac(self, num: int) -> float:
        return num / self.params_total

    @property
    def trainable_frac(self) -> float:
        return self.params_total_trainable / self.params_total


def describe(model: torch.nn.Module) -> SlicetuningStats:
    params_all = list(model.parameters())
    params_total = sum(p.numel() for p in params_all)
    params_total_trainable = sum(p.numel() for p in params_all if p.requires_grad)

    params_tuners_all = list(get_tuners(model))
    params_tuners = sum(p.numel() for p in params_tuners_all)
    params_tuners_trainable = sum(p.numel() for p in params_tuners_all if p.requires_grad)

    biases_all = list(get_bias_params(model))
    params_biases = sum(p.numel() for p in biases_all)
    params_biases_trainable = sum(p.numel() for p in biases_all if p.requires_grad)

    layers_count: Counter[str] = Counter()
    for layer in get_slicetune_layers(model):
        class_name = f"{layer.__class__.__module__}.{layer.__class__.__qualname__}"
        layers_count[class_name] += 1

    return SlicetuningStats(
        params_total,
        params_total_trainable,
        params_tuners,
        params_tuners_trainable,
        params_biases,
        params_biases_trainable,
        layers_count,
    )


def pretty_describe(
    model: torch.nn.Module,
    include_layer_info: bool = True,
    include_param_info: bool = True,
    param_info_brief: bool = False,
    supress_warning: bool = False,
    min_width: int = 50,
) -> str:
    try:
        import prettytable
    except ImportError:
        raise ImportError(
            "prettytable is needed for pretty_describe(). "
            "It comes as an optional dependency slicetune[pretty]"
        )

    stats = describe(model)

    table_params = prettytable.PrettyTable()
    table_params.title = "Slicetuning parameter info".upper().center(min_width)
    table_params.field_names = ["name", "params", "% of total"]
    table_params.align["name"] = "l"
    table_params.align["params"] = "r"
    table_params.align["% of total"] = "r"
    table_params.padding_width = 2

    table_params.add_row(
        [
            "Total".upper(),
            stats._pretty(stats.params_total),
            "",
        ]
    )

    table_params.add_row(
        [
            "|-> trainable",
            stats._pretty(stats.params_total_trainable),
            stats._pretty(stats._frac(stats.params_total_trainable)),
        ]
    )

    table_params.add_row(
        [
            "└-> non-trainable",
            stats._pretty(stats.params_total - stats.params_total_trainable),
            stats._pretty(stats._frac(stats.params_total - stats.params_total_trainable)),
        ]
    )

    if not param_info_brief:

        table_params.add_row(["", "", ""])
        table_params.add_row(["Trainable".upper(), "", ""])

        table_params.add_row(
            [
                "|-> slicetuners",
                stats._pretty(stats.params_tuners_trainable),
                stats._pretty(stats._frac(stats.params_tuners_trainable)),
            ]
        )

        table_params.add_row(
            [
                "|-> biases",
                stats._pretty(stats.params_biases_trainable),
                stats._pretty(stats._frac(stats.params_biases_trainable)),
            ]
        )

        table_params.add_row(
            [
                "└-> others",
                stats._pretty(
                    stats.params_total_trainable
                    - stats.params_tuners_trainable
                    - stats.params_biases_trainable
                ),
                stats._pretty(
                    stats._frac(
                        stats.params_total_trainable
                        - stats.params_tuners_trainable
                        - stats.params_biases_trainable
                    )
                ),
            ]
        )

        table_params.add_row(["", "", ""])
        table_params.add_row(["Non-trainable".upper(), "", ""])

        table_params.add_row(
            [
                "|-> slicetuners",
                stats._pretty(stats.params_tuners - stats.params_tuners_trainable),
                stats._pretty(stats._frac(stats.params_tuners - stats.params_tuners_trainable)),
            ]
        )

        table_params.add_row(
            [
                "|-> biases",
                stats._pretty(stats.params_biases - stats.params_biases_trainable),
                stats._pretty(stats._frac(stats.params_biases - stats.params_biases_trainable)),
            ]
        )

        table_params.add_row(
            [
                "└-> others",
                stats._pretty(stats.params_total - stats.params_tuners - stats.params_biases),
                stats._pretty(
                    stats._frac(stats.params_total - stats.params_tuners - stats.params_biases)
                ),
            ]
        )

    table_params_str = table_params.get_string()

    table_layers = prettytable.PrettyTable()
    table_layers.field_names = ["name", "count"]
    table_layers.header = False
    table_layers.header_style = "upper"
    table_layers.title = "Slicetuning layers".upper().center(min_width)
    table_layers.align["name"] = "l"
    table_layers.align["count"] = "r"
    table_layers.padding_width = 2

    for name, count in stats.layers_count.items():
        table_layers.add_row([name, str(count) + "x"])

    table_layers_str = table_layers.get_string()

    outputs: list[str] = []
    if include_param_info:
        outputs.append(table_params_str)
    if include_layer_info:
        outputs.append(table_layers_str)
    if not supress_warning:
        outputs.append(
            "Note that the number of parameters can be incorrect \n"
            "if you share parameters across modules."
        )

    return "\n\n".join(outputs)



from . import nn, utils
from .model import (
    describe,
    fuse,
    get_slicetune_layers,
    get_tuners,
    mark_for_training,
    pretty_describe,
)

patch_linears = nn.Linear.patch

__all__ = [
    "nn",
    "model",
    "utils",
    "patch_linears",
    "describe",
    "fuse",
    "get_slicetune_layers",
    "get_tuners",
    "pretty_describe",
    "mark_for_training",
]



from __future__ import annotations

import abc
import random
from typing import Any, Callable, Iterable, Literal

import torch
import torch.nn.functional as F
from torch import Tensor
from typing_extensions import Self

import slicetune.utils


class Layer(abc.ABC, torch.nn.Module):
    @classmethod
    @abc.abstractmethod
    def from_standard(cls, layer: torch.nn.Module, **kwargs: Any) -> Self:
        ...

    @classmethod
    @abc.abstractmethod
    def patch(cls, model: torch.nn.Module, **kwargs: Any) -> None:
        ...

    @abc.abstractmethod
    def to_standard(self) -> torch.nn.Module:
        ...

    @abc.abstractmethod
    def get_tuners(self) -> Iterable[torch.nn.Parameter]:
        ...

    @staticmethod
    def get_operation_fn(operation: Literal["add", "mul"]) -> Callable[[Tensor, Tensor], Any]:
        if operation == "add":
            return Tensor.add_
        elif operation == "mul":
            return Tensor.mul_
        else:
            raise ValueError(f"Unknown operation: {operation}")

    @staticmethod
    def get_init_value(operation: Literal["add", "mul"]) -> float:
        if operation == "add":
            return 0.0
        elif operation == "mul":
            return 1.0
        else:
            raise ValueError(f"Unknown operation: {operation}")


class Linear(Layer):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        tuner_size: tuple[int, int] | float | None,
        bias: bool = True,
        operation: Literal["add", "mul"] | None = "add",
        dropout: float = 0.0,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        copy_params_from: torch.nn.Linear | None = None,
        tuner_slice: tuple[slice, slice] | None = None,
        random_generator: random.Random | None = None,
    ) -> None:
        super().__init__()

        self.weight: torch.nn.Parameter
        self.tuner: torch.nn.Parameter
        self.bias: torch.nn.Parameter | None
        self.tuner_dropout: torch.nn.Dropout
        self.in_features: int
        self.out_features: int
        self.operation: Literal["add", "mul"]
        self.operation_fn: Callable[[Tensor, Tensor], Any]
        self.tuner_slice: tuple[slice, slice]

        if operation is None:
            operation = "add"

        if random_generator is None:
            random_generator = random.Random()

        if in_features < 1:
            raise ValueError(f"invalid in_features: {in_features}")

        if out_features < 1:
            raise ValueError(f"invalid out_features: {out_features}")

        if tuner_size is None and tuner_slice is None:
            raise ValueError("Either tuner_size or tuner_slice must be specified")

        if tuner_size is not None and tuner_slice is not None:
            if isinstance(tuner_size, float):
                raise ValueError(
                    "tuner_size as a fraction and tuner_slice cannot be specified together"
                )
            else:
                w, h = tuner_slice
                if tuner_size != (w.stop - w.start, h.stop - h.start):
                    raise ValueError(
                        f"tuner_size and tuner_slice do not match {tuner_size}, {tuner_slice}"
                    )

        if isinstance(tuner_size, float) and not 0 <= tuner_size <= 1:
            raise ValueError(f"invalid tuner_size: {tuner_size}")

        tensor_kwargs: dict[str, Any] = dict(device=device, dtype=dtype)
        weight_shape = out_features, in_features

        if isinstance(tuner_size, float):
            tuner_shape = self._get_random_crop(tuner_size, weight_shape, random_generator)
        elif tuner_size is None:
            assert tuner_slice is not None
            w, h = tuner_slice
            tuner_shape = w.stop - w.start, h.stop - h.start
        else:
            tuner_shape = tuner_size

        if tuner_slice is None:
            tuner_slice = self._get_random_coords(weight_shape, tuner_shape, random_generator)

        if not all(
            [
                0 <= tuner_slice[0].start < tuner_slice[0].stop <= weight_shape[0],
                0 <= tuner_slice[1].start < tuner_slice[1].stop <= weight_shape[1],
                tuner_shape[0] == tuner_slice[0].stop - tuner_slice[0].start,
                tuner_shape[1] == tuner_slice[1].stop - tuner_slice[1].start,
            ]
        ):
            raise ValueError(f"invalid tuner_slice: {tuner_slice}")

        self.in_features = in_features
        self.out_features = out_features
        self.tuner_slice = tuner_slice
        self.operation = operation
        self.operation_fn = self.get_operation_fn(operation)
        self.weight = torch.nn.Parameter(torch.empty(weight_shape, **tensor_kwargs))
        self.tuner = torch.nn.Parameter(torch.empty(tuner_shape, **tensor_kwargs))
        self.tuner_dropout = torch.nn.Dropout(dropout)
        if bias:
            self.bias = torch.nn.Parameter(torch.empty(out_features, **tensor_kwargs))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters(copy_params_from)

    def reset_parameters(self, linear: torch.nn.Linear | None = None) -> None:
        if linear is None:
            # this gives us the same initialization as torch.nn.Linear
            linear = torch.nn.Linear(self.in_features, self.out_features, self.bias is not None)

        with torch.no_grad():
            self.weight[:] = linear.weight
            if self.bias is not None:
                self.bias[:] = linear.bias
            self.tuner.fill_(self.get_init_value(self.operation))

    def forward(self, x: Tensor) -> Tensor:
        return F.linear(x, self._get_w(), self.bias)

    def _get_w(self) -> Tensor:
        w = self.weight.clone()
        tuner = self.tuner_dropout(self.tuner)
        self.operation_fn(w[self.tuner_slice], tuner)
        return w

    @classmethod  # type: ignore
    def from_standard(
        cls,
        layer: torch.nn.Module,
        tuner_size: float | tuple[int, int],
        operation: Literal["add", "mul"] = "add",
        dropout: float = 0.0,
        random_generator: random.Random | None = None,
    ) -> Self:
        """
        Create a slicetune Linear layer from a standard Linear layer.
        By default, the layer will be initialized so that without any training,
        forward() will return the same result as the original Linear layer.
        """
        if not isinstance(layer, torch.nn.Linear):
            raise ValueError(f"Expected linear to be of type torch.nn.Linear, got {type(layer)}")

        return cls(
            in_features=layer.in_features,
            out_features=layer.out_features,
            bias=layer.bias is not None,
            device=layer.weight.device,
            dtype=layer.weight.dtype,
            copy_params_from=layer,
            tuner_size=tuner_size,
            operation=operation,
            dropout=dropout,
            random_generator=random_generator,
        )

    def to_standard(self) -> torch.nn.Linear:
        """
        Converts the slicetune Linear layer to a standard Linear layer, such that
        forward() of the two layers will return the same result (in eval mode)
        """
        was_training = self.training
        self.eval()
        layer = torch.nn.Linear(
            self.in_features,
            self.out_features,
            self.bias is not None,
            device=self.weight.device,
            dtype=self.weight.dtype,
        )
        with torch.no_grad():
            layer.weight[:] = self._get_w()
            if layer.bias is not None:
                assert self.bias is not None
                layer.bias[:] = self.bias
        if was_training:
            self.train()
        return layer

    def get_tuners(self) -> Iterable[torch.nn.Parameter]:
        return [self.tuner]

    @classmethod  # type: ignore
    def patch(
        cls,
        model: torch.nn.Module,
        tuner_size: float | tuple[int, int],
        operation: Literal["add", "mul"] = "add",
        dropout: float = 0.0,
        random_generator: random.Random | None = None,
    ) -> None:
        """
        Replaces all compatible layers in the model with equivalent slicetune layers.
        Disables gradient computation for all parameters except for the slicetuners.
        """
        if isinstance(model, torch.nn.Linear):
            raise ValueError(
                "patch() cannot be called on a torch.nn.Linear. patch() changes model's "
                "layers inplace. For converting a layer, use slicetune.nn.Linear.from_standard() "
            )

        if random_generator is None:
            random_generator = random.Random()

        # prevent invalidating the iterator by setattr
        module_infos = list(slicetune.utils.named_modules_with_parent(model))
        for module_info in module_infos:
            if module_info.parent is None:
                continue
            if isinstance(module_info.module, Layer):
                continue
            if not isinstance(module_info.module, torch.nn.Linear):
                continue

            replaced = Linear.from_standard(
                module_info.module,
                tuner_size=tuner_size,
                operation=operation,
                dropout=dropout,
                random_generator=random_generator,
            )
            setattr(module_info.parent, module_info.attr, replaced)
            module_infos = list(slicetune.utils.named_modules_with_parent(model))

    def __repr__(self) -> str:
        return (
            f"{self.__module__}.{self.__class__.__qualname__}"
            "("
            f"in_features={self.in_features}, "
            f"out_features={self.out_features}, "
            f"bias={self.bias is not None}, "
            f"tuner_size={self.tuner.shape}, "
            f"tuner_slice={self.tuner_slice}, "
            f"operation={self.operation}, "
            f"dropout={self.tuner_dropout.p}"
            ")"
        )

    @staticmethod
    def _get_random_crop(
        fraction: float,
        rect: tuple[int, int],
        random_generator: random.Random,
    ) -> tuple[int, int]:
        a, b = rect
        # We want to have:
        #   c * d ~= fraction * a * b
        # then it holds that:
        #   c ~= fraction * a * b / d
        # if c grows, d must shrink to make up for that area
        #   => to get maximum value for c, plug in d = 1 to the formula above
        #   => to get minimum value for c, plug in d = b
        # this will give us a range of valid values for c
        # now we can sample c from that range and compute d accordingly
        c_max = max(1, min(a, round(fraction * a * b)))
        c_min = max(1, min(a, round(fraction * a)))
        c = random_generator.randint(c_min, c_max)
        d = min(b, max(1, round(fraction * a * b / c)))
        return c, d

    @staticmethod
    def _get_random_coords(
        shape: tuple[int, int],
        crop: tuple[int, int],
        random_generator: random.Random,
    ) -> tuple[slice, slice]:
        start_out = random_generator.randint(0, shape[0] - crop[0])
        start_in = random_generator.randint(0, shape[1] - crop[1])
        return slice(start_out, start_out + crop[0]), slice(start_in, start_in + crop[1])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slicetune']

package_data = \
{'': ['*']}

install_requires = \
['torch>=1.8', 'typing-extensions>=4.4.0,<5.0.0']

extras_require = \
{'pretty': ['prettytable>=3.0.0']}

setup_kwargs = {
    'name': 'slicetune',
    'version': '0.0.1',
    'description': 'A non-invasive parameter-efficient method for finetuning large neural models.',
    'long_description': '# Slicetune 🍕 [WIP, breaking changes expected]\n\nA simple, flexible and non-invasive parameter-efficient method for finetuning large neural models.\n\nWith slicetune you can finetune only a desired fraction of parameters with a few lines of code. Slicetune is compatible with any torch optimizer and with any torch model that uses linear layers, such as (huggingface) transformers. After finetuning, you can "fuse" the trained parameters into the model and get exactly the same architecture as the original checkpoint you started from. You don\'t need to use this package during inference - you don\'t even have to have it installed during inference. \n\n<br/>\n\n\n## Instalation\n\n`pip install slicetune[pretty]`\n\nTODO: publish on pypi\n\noptional dependency `pretty` is for a pretty summary of finetuning parameters.\n\n<br/>\n\n\n## Example usage\n\n```python3\n# create a model\nmodel = transformers.AutoModel.from_pretrained("xlm-roberta-base")\n\n# replace standard layers with equivalent slicetuning layers\nslicetune.patch_linears(model, tuner_size=0.3)\n\n# makes only slicetuners trainable\nslicetune.mark_for_training(model)\n\n# display info about trained params\nprint(slicetune.pretty_describe(model))\n\n# finetune your model\nparams = [p for p in model.parameters() if p.requires_grad]\noptimizer = torch.optim.Adam(params, lr=5e-5)\n...\n\n# replace slicetuning layers with equivalent standard layers\nslicetune.fuse(model)\n\n# Your model is now 100% compatible with the original architecture.\n# You can save it or load it using standard HF methods.\nmodel.save_pretrained("./my_model")\nmodel = transformers.AutoModel.from_pretrained("./my_model")\n```\n\nYou can also use slicetune layers directly when you build a model:\n\n```python3\nINPUT_DIM = 756\nHIDDEN_DIM = 1024\nOUTPUT_DIM = 20\nTUNER_SIZE = 0.2\n\nmodel = torch.nn.Sequential(\n    slicetune.nn.Linear(INPUT_DIM, HIDDEN_DIM, tuner_size=TUNER_SIZE),\n    torch.nn.ReLU(),\n    slicetune.nn.Linear(HIDDEN_DIM, HIDDEN_DIM, tuner_size=TUNER_SIZE),\n    torch.nn.ReLU(),\n    slicetune.nn.Linear(HIDDEN_DIM, HIDDEN_DIM, tuner_size=TUNER_SIZE),\n    torch.nn.ReLU(),\n    slicetune.nn.Linear(HIDDEN_DIM, OUTPUT_DIM, tuner_size=TUNER_SIZE)\n)\n\nslicetune.mark_for_training(model)\n\n```\n\n<br/>\n\n## Q&A\n\n### Why not finetune the whole model?\n\nSlicetuning belongs to a family of *parameter-efficient finetuning methods (PEFT)* that update only a fraction of parameters. There are multiple benefits:\n\n1. **It dramatically decreases (gpu) memory usage.** A full finetuning of a large model requires a lot of memory for optimizer state. In PEFT, optimizer state can be several times smaller.\n1. **It prevents catastrophic forgetting.** Training of all parameters makes models forget general knowledge from pretraining. This is similar to overfitting - it makes the model perform well on finetuning data but badly on unseen data.\n1. **It makes finetuning faster.** Fitting less parameters usually needs less iterations to converge.\n1. **It increases accuracy in low-data setting.** If you have a small training set, you might get a more capable model by training only a part of the model. This is likely because of item 2.\n1. **It increases robustness on out-of-domain data.** This is also likely because of item 2.\n\n\n### How does slicetuning work?\n\nThe idea is to update only selected slices of model parameter tensors. However, torch optimizers can only update whole tensors, not slices. `slicetune` solves it by replacing torch layers with slicetune layers. They contain an extra smaller parameter `tuner` that is added to a slice of `weight` during `.forward()`. Now, you can optimize only the tuner parameters during training. After training, you can swap slicetune layers back with standard torch layers, and "fuse" `tuner` with `weight` and obtain a model with the exact same architecture it had before.\n\n\n### How does slicetuning compare relative to other PEFT methods?\n\nA quantitative comparison is WIP, but we can go through some practical differences.\n\n<details>\n<summary>\n    \n#### Slicetune vs Adapters\n    \n</summary>\n    \n[*adapter-transformers*](https://github.com/adapter-hub/adapter-transformers) is a popular alternative to slicetune. Adapters inject additional low-parameter layers to a model and finetune only those layers. Because they modify the model, a finetuned *Bert* using adapters != *Bert* from *transformers* library. For example, you cannot use standard `transformers` library to load the model using `AutoModel.from_pretrained(\'./your_adapter_checkpoint\')`.\n\nIn order to deal with this, they maintain the library as a fork of *transformers* and try to keep it as compatible as possible while supporting as many features and models as possible. This has some disadvantages - you need to replace your installation of *transformers* with the fork and if it does not support a feature that you used, you are out of luck and need to implement it yourself. In addition, using other libraries or tools that depend on *transformers* can cause problems during 1) installation, because pip will try to download *transformers* and *adapter-transformers*, and 2) during runtime because *adapters-transformers* library overloads tha package name "transformers". This means that when other library runs `import transformers`, it will receive a different package than it expects. During inference, you will also need to use `adapter-transformers` library which can be inconvenient if you want to use *transformers* library in the codebase as well.\n\nIn contrast, *slicetune* allows you to fuse the changes and get back the exact same architecture as it was before. You don\'t need to make any changes to the inference codebase - when you load a model finetuned using *slicetune*, it only differs from the pretrained model in the weight values. It also means that if you can compile the pretrained model using `torch.jit.trace`, you can also compile the finetuned model.\n\n</details>\n<details>\n<summary>\n    \n#### Slicetune vs LoRA\n    \n</summary>\n\n[LoRA](https://github.com/microsoft/LoRA) is another popular alternative. Unlike *adapter-transformers*, it is not implemented as a fork of HF *transformers*, so avoids it avoids the issues mentioned above. LoRA uses a similar approach to slicetune because LoRA layers can be also "fused" standard torch layers, keeping a finetuned model compatible with original model class. A quantitative comparison has not been done yet. \n\n</details>\n<details>\n<summary>\n\n#### Slicetune vs BitFit, Prefix-tuning and finetuning model head\n    \n</summary>\n\nThese PEFT methods are more similar to *slicetune*, because you they don\'t require any changes to architecture.\n    \nIn BitFit, you only update the bias parameters, but you don\'t have a full control about how much you want to update. To give a concrete example, bias parameters are only 0.04% of all parameters in `xlm-roberta-base`. In constract, *slicetune* fives you larger control about how much of the model you want to update - for example `xlm-roberta-base` has 30% of all parameters in linear layers, so you can finetune anywhere from 0 to 30% of all model parameters with `slicetune` (I will consider adding slicetune.nn.Embedding layer in the future). Note that you can easily use BitFit inside *slicetune* library instead of or in addition to slicetuning. See the arguments of `slicetune.mark_for_training()` method.\n\nPrefix-tuning finetunes a fixed-sized sequence of embeddings that you prepend to input embeddings. This influences what the model outputs, so you can treat the prefix as parameters and finetune them. Like BitFit, you don\'t have much control over how many parameters you want to train, because you can\'t make the prefix arbitrarily long. Also, you introduce a computational overhead (because time complexity of self-attention is quadratic w.r.t. sequence length) that will also be present during inference. In addition, you if your model supports input length up to 512 and you use 100 for prefix tuning, you decrease the maximum input size by 20%. You might also need to adjust your inference code for tasks like token classification, because some outputs correspond to artificial prefix instead of input tokens.\n\nFinetuning model head is a simple finetuning method, when only the last part of the model is updated and the "backbone" is viewed as a frozen feature extractor. This method converges quickly but the changes to the model are "shallow". If the features extracted by the backbone are good for the end task, the method will perform well and if they are not, the head might not have enough expressive power to change it. However this method is convenient and does not require any changes to the inference code. I will consider adding some util functions to *slicetune* to support this method and make it more convenient.\n    \nBut what about finetuning several full weight matrices of the model? Well, I would like to compare that to *slicetune* in the future, and if it performs well, I will add some util functions to the library to make it more convenient.\n\n</details>\n\n### Why slicetune layers instead of zeroing-out majority of `.grads` in optimizer before `optimizer.step()`?\n\nBecase optimizing just the tuners inside slicetune layers requires less memory. Let\'s say we have 1024x1024 linear layer and want to update just 256x256 parameters (around 6%). In zeroing-out method, optimizer saves the state for a each 1024x1024 weight matrix. With slicetune layers, the optimizer only saves the state for the small 256x256 matrix.\n\n\n\n<br/>\n',
    'author': 'Marek Kadlcik',
    'author_email': '10684818+markcheeky@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)
